#!/usr/bin/env python
# track a specified event (== program) and record it.
#   temporal suspend/delay of the event is supported
# if run with -c option, service_id arg can be set dummy(>0xFFFF)
#  In this case, the recording timings can be mistaken with those of the
#   another event which shares the same eventID in the TS.

import sys, time
import os
import gobject
import pygst
pygst.require("0.10")
import gst
from optparse import OptionParser

class CLI_Main:

	def __init__(self):
		self.pipeline = gst.Pipeline("pipeline")
		dvb = gst.element_factory_make("dvbbasebin", "dvbbasebin")
#		dvb.set_property("async-handling", True)
		valve = gst.element_factory_make("valve", "valve")
		valve.set_property("drop", True)
		fdsink = gst.element_factory_make("fdsink", "fdsink")
		fdsink.set_property("sync", False)
		self.pipeline.add(dvb, valve, fdsink)
		dvb.get_request_pad("src0")
		gst.element_link_many(dvb, valve, fdsink)
		self.eit_ver_present = None
		self.eit_eid_present = None
		self.eit_ver_following = None
		self.eit_eid_following = None
		self.ev_id = None
		self.ev_start = None
		self.svc_id = None
		self.recording = False
		self.verbose = False
		self.found_ev = False
		self.len = None
		self.loop = None
		bus = self.pipeline.get_bus()
		bus.add_signal_watch()
		bus.connect("message", self.on_message)

	def dprint(self, txt):
		if self.verbose:
			print >> sys.stderr, txt

	def quit(self):
		if self.loop and self.loop.is_running():
			self.loop.quit()

	def on_message(self, bus, message):
		t = message.type
		if t == gst.MESSAGE_EOS:
			print >> sys.stderr, "Finished by unexpected EOS."
			self.quit()
		elif t == gst.MESSAGE_ERROR:
			(err, debug) = message.parse_error()
			print >> sys.stderr, "Error: %s" % err
			self.quit()
		elif t == gst.MESSAGE_ELEMENT:
			self.proc_eit(message)

	def is_target(self, present, tm, dur):
		self.dprint("search %s %s %s..." % (present, tm, dur))
		dur_undef = 165 * 3600 + 165 * 60 + 165 # FF:FF:FF as BCD
		if present:
			# if dur is undefined (extending) and
			#    EIT/f has been already recved: eit_eid_folloing != None
			#      and was not set as target: self.ev_id != eit_eid_following
			#  then set this EIT/p as the target
			#    if EIT/f is not recved yet, then defer the decision
			#    until the next EIT/f  by leaving self.ev_id None for now
			if dur == dur_undef and self.eid_eid_following:
				return True

			# if dur is defined (thus end-time is defined) and
			#   not ending too soon(120sec), then select this EIT/p
			# extra 120sec margin is necessary as this script can/should be
			#  run shortly before the target event starts.
			if dur != dur_undef and time.mktime(tm) + dur > time.time() + 120:
				return True
		else:
			# if this EIT/f is starting very soon, select this EIT/f
			#  but firstly check if start-time is undef (delaying): 1900-0-0-....
			if tm.tm_year != 1900 and time.mktime(tm) < time.time() + 120:
				return True

		return False

	def proc_eit(self, message):
		if message.structure is None:
			return
		eit = message.structure
		if eit.get_name() != "eit":
			return
		if not (eit["actual-transport-stream"] and eit["present-following"]):
			return

		if eit["service-id"] != self.svc_id:
			return

		events = eit["events"]
		if events is None or len(events) < 1:
			return

		tm = time.struct_time([
		  events[0]["year"],
		  events[0]["month"],
		  events[0]["day"],
		  events[0]["hour"],
		  events[0]["minute"],
		  events[0]["second"],
		  0, 1, 0])
		dur = events[0]["duration"]

		if self.ev_id is None and \
		   self.is_target(eit["section-number"] == 0, tm, dur):
			self.ev_id = events[0]["event-id"]
			self.dprint("selected the event:%s..." % self.ev_id)

		if eit["section-number"] == 0:
			if eit["version-number"] == self.eit_ver_present:
				return
			self.eit_ver_present = eit["version-number"]
			self.eit_eid_present = events[0]["event-id"]
		else:
			if eit["version-number"] == self.eit_ver_following:
				return
			self.eit_ver_following = eit["version-number"]
			self.eit_eid_following = events[0]["event-id"]

		self.dprint("EIT section:%d eid:%d svcid:%d ver:%d ev_num:%d" %
				(eit["section-number"], events[0]["event-id"],
				 eit["service-id"], eit["version-number"], len(events)))
		# if event is still not identified, then return & wait subsequent EITs
		if self.ev_id is None:
			return

		if self.ev_id == self.eit_eid_present:
			self.found_ev = True
			if self.ev_start is None:
				self.ev_start = time.mktime(tm)
			self.valve(False)
			return
		elif self.ev_id == self.eit_eid_following:
			self.found_ev = True
			if self.ev_start is None and tm.tm_year != 1900:
				self.ev_start = time.mktime(tm)
			# check if the next event is close enough
			#  but exclude the case for the starting period of unplanned pause.
			#   in this case, the start time is set to the past.
			# (assert the updated EIT is the "following" one)
			now = time.time()
			if eit["section-number"] == 1 and \
			   self.ev_start >= now and self.ev_start - 5 <= now:
				self.valve(False)
			else:
				self.valve(True)
			return
		else:
			# check if in transient state for unplanned pause/break-ins
			if self.recording and eit["section-number"] == 0 and \
			   self.eit_ver_present != self.eit_ver_following:
				self.valve(True)
				return
			# check if in state for initial waiting
			if not self.found_ev:
				return
			self.valve(True)
			self.dprint("Finished recording.")
			self.quit()
			return

	def valve(self, drop):
		if self.recording != drop:
			return
		self.pipeline.get_by_name("valve").set_property("drop", drop)
		self.recording = not drop
		if drop:
			self.dprint("Paused recording...")
		else:
			self.dprint("Start recording evend:%d..." % self.ev_id)
			if self.len is not None and self.len > 0:
				gobject.timeout_add_seconds(60 * self.len, self.on_stop)
				self.len = None

	def on_stop(self):
		self.valve(True)
		self.dprint("Finished recording by timer.")
		self.quit()
		return False

	def on_timer(self):
		if self.found_ev:
			return False
		self.dprint("Cannot find the event. giving up...")
		self.quit()
		return False

	def set_timer(self, wait):
		if wait is None or wait <= 0:
			wait = 30
		gobject.timeout_add_seconds(wait * 60, self.on_timer)

#
# helper func.
#
def check_channel(chname, progid, confname):
	if confname is None:
		if os.environ.has_key('GST_DVB_CHANNELS_CONF'):
			confname = os.environ['GST_DVB_CHANNELS_CONF']
		else:
			confname = os.path.join(os.environ['HOME'],
				".gstreamer-%d.%d" % gst.version()[0:2],
				"dvb-channels.conf")

	with open(confname) as f:
		for line in f:
			if line.startswith('#'):
				continue
			items = line.strip().split(':')
			if len(items) < 3:
				continue
			if chname and chname != items[0]:
				continue
			elif progid and progid != int(items[-1], 0):
				continue
			return (items[0], int(items[-1], 0));

	return (None, None)

#
# start of the main()
#
gobject.threads_init()
usage = "usage: %prog [options] {-c CH_NAME | -s SERVICE_ID}"

parser = OptionParser(usage=usage)
parser.add_option("-v", "--verbose",
					action="store_true",
					help="output informational messages. [default: False]")

parser.add_option("-c", "--channel",
					metavar="NAME",
					help="channel name of the recording event. "
							" [default: derived from service ID & conf-file]")
parser.add_option("-s", "--serviceid",
					metavar="ID", type="int",
					help="service_ID (program_ID) of the recording event. "
							" [default: derived from channel name & conf-file]")

parser.add_option("-e", "--eventid",
					metavar="ID", type="int",
					help="event_ID which identifies the recording event. "
							" [default: the current event on air]")

parser.add_option("-w", "--wait",
					metavar="TIME", type="int", default=30,
					help="max wait (in min.) until the event is found in EIT."
							" [default: %default min.]")

parser.add_option("-l", "--length",
					metavar="TIME", type="int",
					help="recording length (in min.) [default: whole event]")

parser.add_option("-o", "--output",
					dest="filename", metavar="FILE",
					help="write output to FILE. [default: stdout]")

parser.add_option("-a", "--adapter",
					metavar="NUM", type="int",
					help="use DVB adapter #NUM. [default: 0]")
#parser.add_option("-d", "--demuxer",
#					metavar="NUM", type="int",
#					help="use DVB demuxer #NUM. [default: 0]")
parser.add_option("-f", "--frontend",
					metavar="NUM", type="int",
					help="use DVB frontend #NUM. [default: 0]")

parser.add_option("--conf",
					metavar="FILE",
					help="path to the channel config file. "
							" [default:~/.gstreamer-%d.%d/dvb-channels.conf]"
							 % gst.version()[0:2])

(options, args) = parser.parse_args()

try:
	mainclass = CLI_Main()

	if options.channel is None and options.serviceid is None:
		parser.error("at least either -c or -s option must be specified.")

	mainclass.verbose = options.verbose
	mainclass.len = options.length

	if options.filename:
		try:
			f = open(options.filename, 'w')
			mainclass.pipeline.get_by_name("fdsink").set_property("fd", f.fileno())
		except:
			parser.error("option -o has unwritable filename: %s" % options.filename)

	if options.adapter:
		mainclass.pipeline.get_by_name("dvbbasebin").set_property("adapter", options.adapter)
	if options.frontend:
		mainclass.pipeline.get_by_name("dvbbasebin").set_property("frontend", options.frontend)

	if options.eventid:
		ev_id = options.eventid
	else:
		ev_id = None
	if options.eventid and ev_id < 0 or ev_id >= 0x10000:
		parser.error("invalid event_id:%d." % ev_id)
	mainclass.ev_id = ev_id

	(ch_name, sid) = check_channel(options.channel, options.serviceid, options.conf)
	if ch_name is None or sid is None:
		parser.error("(channel:%s, service_id:%s) is not a valid combination."
			% (options.channel, options.serviceid))
	mainclass.svc_id = sid
	mainclass.pipeline.get_by_name("dvbbasebin").set_uri("dvb://%s" % ch_name)

	loop = gobject.MainLoop()
	mainclass.loop = loop
	mainclass.pipeline.set_state(gst.STATE_PLAYING)
	mainclass.dprint("Waiting for the EIT (ID:%s) to be received..." % ev_id)
	mainclass.set_timer(options.wait)
	loop.run()
except:
	if mainclass:
		mainclass.quit()
